import datetime
import json
import logging
from collections import UserDict, UserList
from datetime import datetime

from microservice.exceptions import ApiError, CriticalError


class SerializableMixin:
    def list(self, l=None):
        """return raw list"""
        raw_list = []
        for v in l if l is not None else self:
            if isinstance(v, BasicObject):
                raw_list.append(v.dict(v))
            elif isinstance(v, Collection):
                raw_list.append(v.list(v))
            elif isinstance(v, dict):
                raw_list.append(self.dict(v))
            elif isinstance(v, list):
                raw_list.append(self.list(v))
            else:
                raw_list.append(v)
        return raw_list

    def dict(self, d=None):
        """return raw dict"""
        raw_dict = {}
        for k, v in d.items() if d is not None else self.items():
            if isinstance(v, BasicObject):
                raw_dict[k] = v.dict(v)
            elif isinstance(v, Collection):
                raw_dict[k] = v.list(v)
            elif isinstance(v, dict):
                raw_dict[k] = self.dict(v)
            elif isinstance(v, list):
                raw_dict[k] = self.list(v)
            elif isinstance(v, datetime):
                raw_dict[k] = v.isoformat("T")
            else:
                raw_dict[k] = v
        return raw_dict

    def json(self):
        if isinstance(self, BasicObject):
            return json.dumps(self.dict())
        elif isinstance(self, Collection):
            return json.dumps(self.list())

    @staticmethod
    def from_json(s):
        o = json.loads(s)
        if isinstance(o, list):
            return Collection(o)
        elif isinstance(o, dict):
            return BasicObject(o)
        else:
            return o

    def __repr__(self):
        return self.json()

    def __str__(self):
        return self.json()


class BasicObject(UserDict, SerializableMixin):
    def __init__(self, item_dict: dict, obj=None):
        super(BasicObject, self).__init__()  # empty until validated
        for key in item_dict:  # protect system fields
            if key in ["data"]:
                item_dict[key + "_"] = item_dict[key]
                del item_dict[key]

        self.input = item_dict
        self.validate()
        # after this moment self.data is validated and may be converted to object fields
        for key in self.data:
            if isinstance(self.data[key], dict):
                self.data[key] = BasicObject(self.data[key])
                # logging.warning("Nested object without class {}: {}".format(self.__class__.__name__, key))
            elif isinstance(self.data[key], list):
                self.data[key] = Collection(self.data[key])
                # logging.warning("Nested list without class {}: {}".format(self.__class__.__name__, key))
            setattr(self, key, self.data[key])
        self.model = None
        self.obj = obj

    def __setitem__(self, key, value):
        self.data[key] = value
        setattr(self, key, value)

    def __delitem__(self, key):
        del self.data[key]
        delattr(self, key)

    def validate(self):
        """implement it otherwise self.data will be filled from self.input directly"""
        self.data = self.input

    def valid(self, name, default=None, coerce=None, check=None, error="", required=False, allow_none=False):
        """
        Validate field and coerce it if needed. Each call fill self.data with new field.
        :param name: field name
        :param default: default value if name not presents in data
        :param coerce: coerce to type of function
        :param check: check function must return True if passed
        :param error: error text if any of checks not passed
        :param required: set True if field is required
        :param allow_none: set True if param may be None, else None values will be removed
        """
        value = self.input.get(name, default)
        if required and (value in (None, "") or name not in self.input):
            raise ApiError("#missing #field '{}' {}".format(name, error))
        if name not in self.input and default is None:
            return
        if value is None:
            value = default
        if coerce and value is not None:
            try:
                value = coerce(value)
            except ValueError:
                raise ApiError("#wrong_type {}, expected {}. {}".format(name, coerce, error))
        if check and value is not None:
            if isinstance(check, (list, tuple)) and value not in check:
                raise ApiError("#wrong_format {} not in {}. {}".format(name, check, error))
            if callable(check) and not check(value):
                raise ApiError("#wrong_format {}. {}".format(name, error))
        # final
        if value is not None or allow_none:
            self.data[name] = value
            return value

    def set_model(self, model, obj=None):
        """if object have id, it may be saved to db (o rly?)"""
        self.model = model
        self.obj = obj or self.obj
        return self

    async def save(self):
        if not self.model or not self.obj:
            raise CriticalError("Model or objects manager is not set")
        if not self.get("id"):
            raise CriticalError("Object id is not set")
        model = self.model.from_dict(self.dict())
        await self.obj.update(model)

    def group_enum(self, enum_name, variants):
        """
        Group several boolean flags into one enum field
        :param enum_name: name of new field
        :param variants: flags
        """
        for var in variants:
            if getattr(self, var):
                self[enum_name] = var
                break

    def __sub__(self, other):
        """return dict with fields which is differs in self and other"""
        res = BasicObject({})
        for key in self:
            if not other.get(key) or self[key] != other.get(key):
                res[key] = self[key]
        return res

    def drop_nones(self, required=None):
        clear = {}
        for k, v in self.items():
            if v is not None or (required and k in required):
                clear[k] = v
            else:
                delattr(self, k)
        self.data = BasicObject(clear)


class Collection(UserList, SerializableMixin):
    object_class = None

    def __init__(self, items_list, object_class=None):
        self.object_class = object_class or self.object_class
        super(Collection, self).__init__(items_list)
        if self.object_class:
            self.set_class(self.object_class)
            self.valid = True
        else:
            self.valid = False
            self.set_class(BasicObject)
        self._ix = {}

    def set_class(self, object_class):
        self.object_class = object_class
        if all([isinstance(x, dict) for x in self.data]):
            self.data = [object_class(item) for item in self.data]

    @property
    def ix(self):
        """
        Need id of items
        """
        if not self._ix and len(self) > 0 and self[0].get("id"):
            self._ix = {v["id"]: v for v in self}
        return self._ix

    def join(self, children, foreign_key, group_name, how="many"):
        assert how in ("many", "one")
        for i in self.data:
            i[group_name] = [] if how == "many" else None
        for child in children:
            try:
                if how == "many":
                    self.ix[child[foreign_key]][group_name].append(child)
                elif how == "one":
                    self.ix[child[foreign_key]][group_name] = child
            except KeyError:
                pass
                # logging.debug("Lost object index: {} id {}".format(foreign_key, child[foreign_key]))

    @classmethod
    def with_class(cls, object_class_=None):
        """Return new class"""
        class TypedCollection(Collection):
            object_class = object_class_ or BasicObject
        logging.debug("Redefined class in collection: {} (Collection id {}), original: {} (Collection id {})".format(TypedCollection.object_class, id(TypedCollection), cls.object_class, id(cls)))
        return TypedCollection

    def group_enum(self, enum_name, variants):
        """
        Group several boolean flags into one enum field
        :param enum_name: str name of new field
        :param variants: list flags
        """
        for item in self.data:
            item.group_enum(enum_name, variants)

    def search(self, case_insensitive=True, method="and", **kwargs):
        """
        Search and return filtered collection with linked objects
        :param case_insensitive: only if value is str
        :param method: and/or
        :param kwargs: filters
        :return: Collection
        """
        results = Collection([])
        for item in self:
            match = 0
            for key, value in kwargs.items():
                if item[key] == value or (case_insensitive and isinstance(item[key], str) and str(item[key]).lower() == str(value).lower()):
                    match += 1
            if method == "and" and match == len(kwargs):
                results.append(item)
            elif method == "or" and match > 0:
                results.append(item)
        return results


if __name__ == '__main__':
    a = Collection([])
    a.append({"a": "a", "b": "b"})
    a.append({"a": 1, "b": "2"})
    a.append({"a": "a", "b": "b"})
    print(a.search(a=1, b="2", method="and"))



import json

from microservice.middleware.objects import BasicObject


class RedisCache:
    def __init__(self, redis_connection):
        self.redis = redis_connection

    async def store_request(self, request_hash, expire, value):
        data = json.dumps(value)
        await self.redis.setex("request.{}".format(request_hash), expire, data)

    async def restore_answer(self, request_hash):
        data = await self.redis.get("request.{}".format(request_hash))
        answer = json.loads(data)
        return answer

    async def check_request(self, request_hash):
        return await self.redis.exists("request.{}".format(request_hash))

    async def set(self, key, expire=0, value=None):
        data = json.dumps(value)
        await self.redis.setex(key, expire, data)

    async def get(self, key):
        data = await self.redis.get(key)
        value = json.loads(data)
        return value

    async def store(self, key, expire=0, o=None):
        """
        Store BasicObject to redis (jsoned)
        :param key: 
        :param expire: 
        :param o: BasicObject
        :return: None
        """
        if o:
            data = o.json()
        else:
            data = None
        await self.redis.setex(key, expire, data)

    async def restore(self, key, object_class=BasicObject):
        """
        Restore object from redis
        :param key:
        :param object_class:
        :return: BascicObject
        """
        value = await self.redis.get(key)
        o = object_class.from_json(value)
        return o

    async def exists(self, key):
        return await self.redis.exists(key)

    async def list(self, prefix):
        keys = await self.redis.keys("{}*".format(prefix))
        items = []
        if keys:
            items = [BasicObject.from_json(x) for x in await self.redis.mget(*keys)]
        return items

    async def delete(self, key):
        await self.redis.delete(key)



import asyncio
import logging
from traceback import print_exc

import cfg
import peewee_async
from raven.contrib.tornado import AsyncSentryClient
from tornado.platform.asyncio import AsyncIOMainLoop
from tornado.web import Application

from microservice.routes import Router


class Server:
    def __init__(self, handlers):
        logging.info("Server starting...")
        AsyncIOMainLoop().install()
        self.loop = asyncio.get_event_loop()

        self.router = Router(handlers)
        self.app = Application(self.router.routes, **cfg.app.tornado_settings)
        self.app.loop = self.loop

        if hasattr(cfg, "db"):
            from microservice.models import connection
            self.app.objects = peewee_async.Manager(connection, loop=self.loop)
        else:
            self.app.objects = None

        self.app.sentry_client = AsyncSentryClient(cfg.app.sentry_url)
        try:
            import aioredis
            if hasattr(cfg.redis, "socket") and cfg.redis.socket:
                self.app.redis_connection = self.loop.run_until_complete(aioredis.create_redis_pool(cfg.redis.socket))
            else:
                self.app.redis_connection = self.loop.run_until_complete(aioredis.create_redis_pool((cfg.redis.host, cfg.redis.port)))
        except (ImportError, OSError, AttributeError):
            self.app.redis_connection = None

    def run(self):
        try:
            self.app.listen(port=cfg.app.port, address="0.0.0.0")
            logging.info("Server started on port: {}, debug mode: {}".format(cfg.app.port, cfg.app.debug))
            self.loop.run_forever()
        except KeyboardInterrupt:
            self.loop.close()
            logging.info("Server stopped")
        except:
            print_exc()
            self.app.sentry_client.captureException()
            logging.error("Server crashed!")
            exit(1)



class ApiError(BaseException):
    pass


class CurvedHands(BaseException):
    pass


class InternalError(BaseException):
    pass


class AccessDenied(BaseException):
    pass


class CriticalError(BaseException):
    pass



import logging
import os
import time
from logging.config import dictConfig


class redis:
    host = os.getenv("REDIS_HOST", "localhost")
    port = os.getenv("REDIS_PORT", "6379")


class app:
    debug = bool(os.getenv("DEBUG", True))
    port = 8001
    workdir = os.path.dirname(os.path.abspath(__file__))
    domain = "example.com"
    secret_key = "random string"
    session_lifetime = 90  # days

    timezone = "UTC"  # server timezone

    gzip_output = 0  # compress level; 0 - off, 9 - max

    tornado_settings = {
        "debug": debug,
        "compress_response": True if gzip_output > 0 else False,
    }

    smtp_host = "example.com"
    smtp_email = "contact@example.com"
    smtp_login = "contact@example.com"
    smtp_password = "password"
    smtp_from = "Mailer <contact@example.com>"

    support_email = "contact@example.com"
    send_email_errors = False
    errors_email = "contact@example.com"

    sentry_url = os.getenv("SENTRY_URL")
    send_telegram_errors = True
    telegram_reporter = {
        "unknown": {  # hostname
            "chat_id": "-123",
            "name": "Unknown Host",
        },
    }
    telegram_bot_token = "123:lalala"

    logging_config = {
        "version": 1,
        "formatters": {
            "basic": {
                "format": "%(asctime)s %(name)-16s %(levelname)-8s %(message)s"
            }
        },
        "handlers": {
            "stream": {
                "class": "logging.StreamHandler",
                "formatter": "basic",
                "level": logging.INFO
            }
        },
        "root": {
            "handlers": ["stream"],
            "level": logging.INFO
        },
    }

    for handler in ["tornado", "botocore", "boto3"]:  # ignored logging users
        logging.getLogger(handler).setLevel(logging.ERROR)
    dictConfig(logging_config)
    os.environ["TZ"] = timezone
    time.tzset()

    number_of_nodes = 1
    max_workers_on_executor = 16
    ipstack_api_key = ""


class Enum:
    @classmethod
    def all(cls):
        return [cls.__dict__[x] for x in cls.__dict__.keys() if not x.startswith("_")]



import logging
from asyncio import iscoroutine
from collections import UserDict
from datetime import datetime, timezone, date, time, timedelta
from uuid import UUID

from microservice.middleware.objects import Collection, BasicObject
from microservice.exceptions import CurvedHands


class Data:
    """
    Контейнер данных, возвращаемый менеджером. Приводит результат к словарю для сериализации.
    Обязательно должен содержать result или error.
    """
    def __init__(self, result=None, error=None, meta=None, need_format=True, reformat=True, transpose=None, **kwargs):
        """
        Создать контейнер на основе объекта Model
        :param result: результат запроса peewee (dict or list)
        :param error: текст ошибки
        :param meta: метаданные от менеджера
        :param need_format: нужно применить format
        :param reformat: применить метод reformat
        :param transpose: преобразовать в словарь (транспонированием)
        """
        assert result is not None or error is not None
        self.kwargs = kwargs
        self.meta = dict()
        self.error = error
        if not error:
            if isinstance(result, (Collection, list)):
                # list of objects
                if need_format:
                    self.data = list(map(self.format, result))
                else:
                    self.data = result
                if reformat:
                    self.data = list(map(self.reformat, self.data))
                if transpose:
                    self.transpose(transpose[0], transpose[1])
                self.count = len(self.data)
                self.meta["shown"] = self.count
                self.data = self._cast(self.data)
            elif isinstance(result, (BasicObject, UserDict, dict)):
                # single object
                if need_format:
                    self.data = self.format(result)
                else:
                    self.data = result
                if reformat:
                    self.data = self.reformat(self.data)
                self.count = 1
                self.meta["shown"] = self.count
                self.data = self._cast(self.data)
            elif iscoroutine(result):
                raise CurvedHands("You forgot 'await' statement")
            else:
                self.data = result
                logging.debug("Data is in unknown format: {}".format(self.data))
                self.error = "Data is in unknown format"
            if meta:
                self.meta.update(meta)
        else:
            self.data = None

    @staticmethod
    def format(item):
        """
        Задать интерфейс объекта. Метод для переопределния.
        :param item: dict соответствующий модели
        :return: dict для вывода (эластик)
        """
        return item

    @staticmethod
    def reformat(item):
        """
        Выполнить переформатирование готового объекта
        :param item: dict для вывода (эластик)
        :return: dict измененный
        """
        return item

    @classmethod
    def list(cls, items):
        """
        Сделать список используя self.format. Используется в дочерних классах, пример:
            def format(item):
                item = {
                    "users": ContributorsData.list(item["users"]),
                    "users_count": item["users_count"],
                }
                return item
        :param items: список для форматирования
        :return: форматированный список
        """
        if items is not None:
            items = list(map(cls.format, items))
        else:
            items = []
        return items

    def _cast(self, value):
        """
        Приводит значение к виду удобному для сериализации
        Если значение является словарем или списком, преобразует рекурсивно
        """
        # преобразовать определенные типы (добавить нужное)
        if isinstance(value, (datetime, time)):
            return value.replace(tzinfo=timezone.utc).isoformat()
        elif isinstance(value, date):
            return value.replace().isoformat()
        elif isinstance(value, timedelta):
            return value.days
        elif isinstance(value, UUID):
            return str(value)
        elif iscoroutine(value):
            raise CurvedHands("You forgot 'await' statement")

        # рекурсивно пройтись по элементам списка или словаря
        elif isinstance(value, dict):
            return {k: self._cast(v) for k, v in value.items()}
        elif isinstance(value, list):
            return [self._cast(v) for v in value]
        elif isinstance(value, BasicObject):
            return self._cast(value.dict())
        elif isinstance(value, Collection):
            return self._cast(value.list())
        else:
            return value

    def transpose(self, key, value=None):
        """
        Превратить список элементов в словарь с ключом key (одно из полей)
        :param key: поле для использования в качестве ключа, должно быть уникальным, иначе будет потеря данных
        :param value: оставить только это поле в качестве значения элементов словаря. Если не указано, то значение
            словаря это словарь со всеми остальными полями
        """
        if value:
            self.data = {x[key]: x[value] for x in self.data}
        else:
            self.data = {x[key]: x for x in self.data}

    def group_by(self, field, objects=False, group_name="items"):
        """
        Превратить список элементов в словарь, в котором элементы сгруппированы по полю field, аналог group by в sql
        ключом является значение поля field
        значением является список элементов, у которых было определенное поле field
        количество групп - количество уникальных значений field
        :param field: поле для группировки
        :param objects: создать объекты, а не строки
        :param group_name: имя для списка элементов в группе
        """
        groups = {}
        for item in self.data:
            if not item[field] in groups:
                groups[item[field]] = []
            groups[item[field]].append(item)
        if objects:
            groups = [{field: k, group_name: v} for k, v in groups.items()]
        self.data = groups

    def __str__(self):
        return "<{} {}>".format(self.__class__.__name__, self.data)

    @staticmethod
    def extract_field(item, field):
        """
        Extract tuple (id, object) from field if possible
        :param item:
        :param field: field name
        :return: (id, object)
        """
        if isinstance(item.get(field), BasicObject):
            return item.get(field).id, item.get(field)
        elif isinstance(item.get(field), int):
            return item.get(field), None
        elif isinstance(item.get("{}_id".format(field)), BasicObject):
            return item.get("{}_id".format(field)).id, item.get("{}_id".format(field))
        elif isinstance(item.get("{}_id".format(field)), int):
            return item.get("{}_id".format(field)), None
        else:
            return item.get(field), item.get(field)



from prometheus_client import generate_latest
from tornado import gen

from microservice.handlers.handler import BasicHandler


class MetricsHandler(BasicHandler):
    def prepare(self, *kwargs):
        pass

    async def get(self):
        self.finish(generate_latest())

    @gen.coroutine
    def on_finish(self):
        pass



#!/usr/bin/env python3
import logging
import pkgutil
import sys

import migrations
from microservice.models import connection
from models import Migrations


def migrate():
    """
    Функция для проверки и применения миграций.
    """
    logging.info("Trying migrations")
    applied_migrations = [x.name for x in Migrations.select().order_by(Migrations.name).asc()]
    last_migration = applied_migrations[-1] if len(applied_migrations) > 0 else "Database is empty"
    for loader, name, _ in pkgutil.walk_packages(migrations.__path__):
        if name not in applied_migrations:
            Migration = loader.find_module(name).load_module(name).Migration

            # separate migration
            connection.execute_sql("SET STATEMENT_TIMEOUT TO 60000;")
            with connection.atomic():
                try:
                    logging.info("Applying: {}".format(name))
                    Migration()
                except Exception as e:
                    logging.error("Failed on step: {}. "
                                  "Rolling back because of error: {}".format(name, e))
                    connection.rollback()
                    exit(1)
                else:
                    last_migration = name
                    m = Migrations(name= name)
                    m.save()

    logging.info("Last migration is: {}".format(last_migration))


if __name__ == '__main__':
    _, func, *args = sys.argv
    globals()[func](*args)
    logging.info("Done")
    # TODO: переписать с использованием argparse



from datetime import timedelta, datetime

from microservice.middleware.objects import BasicObject, Collection


class DataManager:
    """
    Менеджер данных выполняет работу с данными в формате CRUD+L(list)
    """
    model = None

    def __init__(self, obj=None):
        """        
        Создать менеджер данных
        """
        self.cache = {}
        if obj is not None:
            self.obj = obj
    
    def create(self, **kwargs):
        return None
    
    def read(self, **kwargs):
        return None
    
    def update(self, **kwargs):
        return None
    
    def delete(self, **kwargs):
        return None
    
    def list(self, **kwargs):
        return None

    def save_cache(self, user, name, value, lifetime=timedelta(days=1)):
        key = "{}-{}".format(user["id"], name)
        deathtime = datetime.now() + lifetime
        self.cache[key] = [deathtime, value]

    def load_cache(self, user, name):
        key = "{}-{}".format(user["id"], name)
        if key in self.cache and self.cache[key][0] > datetime.now():
            return self.cache[key][1]
        else:
            del self.cache[key]

    def check_cache(self, user, name):
        key = "{}-{}".format(user["id"], name)
        if key in self.cache and self.cache[key][0] > datetime.now():
            return True
        else:
            return False

    def __del__(self):
        self.cache = {k: v for k, v in self.cache.items() if v[0] > datetime.now()}

    @staticmethod
    def extract_results(results, *args, **kwargs):
        """
        Извлечь список из результатов пиви или эластика
        :param extra_attrs: аттрибуты из результата, которые надо включить в словарь (peewee)
        :param results: ответ эластика или пиви
        :return: list
        """
        if len(results) > 0:
            lst = Collection([x.dict(*args, **kwargs) for x in results], BasicObject)
        else:
            lst = Collection([], BasicObject)
        return lst

    @staticmethod
    def extract_one(results, *args, **kwargs):
        for x in results:
            return BasicObject(x.dict(*args, **kwargs))
        return None



from datetime import datetime, timedelta
from hashlib import sha1
from os import urandom

from peewee import DoesNotExist
from ua_parser import user_agent_parser

import cfg
from microservice.managers.manager import DataManager
from microservice.models import Session


class SessionManager(DataManager):
    user_id_key = "id"
    model = Session

    async def create(self, user, user_agent, network=None, push_token=None, ip=None, location=None):
        ua_data = user_agent_parser.Parse(user_agent)
        os = ua_data["os"]["family"]
        client = ua_data["user_agent"]["family"]
        rand = sha1(urandom(16)).hexdigest()

        session_obj, created = await self.obj.get_or_create(
            self.model,
            user=user["id"],
            user_agent=user_agent,
            login_method=network,
            push_token=push_token,
            defaults=dict(
                os=os,
                client=client,
                rand=rand,
                token=self._generate_token(user[self.user_id_key], os, client, rand),
                expire=datetime.now() + timedelta(days=cfg.app.session_lifetime),
                ip=ip,
                location=location
            )
        )
        session_obj.ip = ip
        session_obj.location = location
        session_obj.last_login = datetime.now()
        session_obj.expire = datetime.now() + timedelta(days=cfg.app.session_lifetime)
        await self.obj.update(session_obj)
        session = session_obj.dict()

        return session

    def read(self, token):
        try:
            if not token:
                raise DoesNotExist
            session_obj = self.model.get(token=token)
            if datetime.now() > session_obj.expire:
                session = None
            else:
                session = session_obj.dict()
        except DoesNotExist:
            session = None
        return session

    async def delete(self, token):
        try:
            if not token:
                raise DoesNotExist
            session_obj = await self.obj.get(self.model, token=token)
            await self.obj.delete(session_obj)
            session = None
        except DoesNotExist:
            session = None
        return session

    async def update(self, token, ip=None, location=None, push_token=None):
        session_obj = await self.obj.get(self.model, token=token)
        session_obj.expire = datetime.now() + timedelta(days=cfg.app.session_lifetime)
        session_obj.ip = ip
        session_obj.location = location
        session_obj.last_login = datetime.now()
        if push_token:
            session_obj.push_token = push_token
        await self.obj.update(session_obj)
        return session_obj.dict()

    def _generate_token(self, user_id, os, client, rand):
        """
        Сгенерировать токен из данных сессии
        :param user_id: id юзера
        :param os: тип ОС
        :param client: тип клиента (браузер)
        :param rand: соль
        :return: строка токена
        """
        # строка для составления хеша токена
        string = str(user_id) + os + client + rand + cfg.app.secret_key
        token = sha1(string.encode("utf-8")).hexdigest()
        return token



from peewee import DoesNotExist

from microservice.managers.manager import DataManager
from microservice.models import Social


class SocialManager(DataManager):
    model = Social

    async def create(self, user, social_data):
        social_obj, created = await self.obj.get_or_create(
            self.model,
            network=social_data["reg_method"],
            social_id=social_data.get("social_id"),
            defaults={
                "access_token": social_data.get("access_token"),
                "user": user["id"]
            }
        )
        social = social_obj.object()
        return social

    async def read(self, network, social_id):
        if not social_id:
            return None
        try:
            res = await self.obj.get(Social, network=network, social_id=social_id)
        except DoesNotExist:
            return None
        else:
            social = res.object()
            return social

    async def delete(self, user, network):
        await self.obj.execute(Social.raw("""
            DELETE FROM social
            WHERE user_id = %s and network = %s
            RETURNING 1
        """, user.id, network))



import logging
from concurrent.futures import CancelledError
from traceback import print_exc

import peewee
import psycopg2
from psycopg2.extensions import QueryCanceledError

from microservice.exceptions import ApiError, InternalError, AccessDenied
from microservice.functions import check_atomic
from microservice.functions import location
from microservice.middleware.user import BaseUser


def check(anonymous=True, roles=None):
    """
    Обертка для использования параметров авторизации
    :param anonymous: разрешить досутп анонимусам (для получения инфы, если юзер залогинен)
    Использование для методов BasicHandler:

    @check_self()
    async def method():
        ...
    """
    def check_auth(method):
        """
        Декоратор для проверки авторизации.
        добавляет параметр me (dict) с данными юзера
        """
        async def wrap(self, *args, **kwargs):
            """
            Выполняет метод, как атомик вызов обернутый в таск
            """
            if not self.cached:
                async def task(self, *args, **kwargs):
                    async with check_atomic(self.application.objects, self.atomize):
                        try:
                            me = await self.session.me()
                            if not (me or anonymous):
                                raise AccessDenied("#auth #token Token invalid")
                            elif roles and me["role"] not in roles:
                                raise AccessDenied("#auth You shall not pass")
                            else:
                                anonymous or await me.check_self(self)
                                if me:
                                    me["sentry_client"] = self.application.sentry_client
                                    me["api_version"] = self.api_version
                                await method(self, me, *args, **kwargs)
                            if self.cache is not None and self.cache_method is not None:
                                await self.cache.store_request(self.request_hash(for_user=self.cache_method == "user"), self.cache_lifetime, self.answer.answer)
                        except InternalError as e:
                            self.compose(error=str(e), send=True)
                        except ApiError as e:
                            self.compose(error="#api_error {}".format(e), send=True)
                        except AccessDenied as e:
                            self.compose(error="#access_denied {}".format(e), status=401, send=True)
                        except (QueryCanceledError, CancelledError) as e:
                            self.captureException()
                            self.compose(error="#db_error query cancelled: {}".format(e), status=500, send=True)
                        except Exception as e:
                            print_exc()
                            raise e
                        else:
                            self.send_result()
                try:  # handle db errors
                    return await self.loop.create_task(task(self, *args, **kwargs))
                except (peewee.OperationalError, psycopg2.OperationalError, peewee.InternalError, peewee.InterfaceError) as e:
                    print_exc()
                    self.captureException()
                    self.on_finish()
                    exit(1)

        return wrap
    return check_auth


class Session:
    """
    Авторизует пользователя, сохраняет сессию, контролирует доступ к данным
    Есть доступ ко всем данным запроса
    """
    usermanager_class = None
    sessionmanager_class = None
    auth_header = "X-Token"
    base_user_class = BaseUser

    def get_usermanager_class(self):
        from microservice.managers.user import UserManager
        return self.usermanager_class or UserManager

    def get_sessionmanager_class(self):
        from microservice.managers.session import SessionManager
        return self.sessionmanager_class or SessionManager

    def __init__(self, request, body, args, obj):
        """
        Создать обработчик сессии
        :param request: объект с параметрами запроса
        :param body: функция досутпа к body
        :param args: функция доступа к параметрам url
        :return: возвращает словари
        """
        self.obj = obj
        self.request = request
        self.args = args
        self.body = body
        self.user = None
        if obj:
            self.user_manager = self.get_usermanager_class()(self.obj)
            self.session_manager = self.get_sessionmanager_class()(self.obj)

    async def force_login(self, user, network=None, push_token=None):
        user_agent = self.request.headers.get("User-Agent", "")
        ip = self.request.remote_ip
        country = await location(ip)
        session_result = await self.session_manager.create(user, user_agent, network, push_token, ip=ip, location=country)
        user["token"] = session_result["token"]
        return user

    async def logout(self):
        token = self.request.headers.get(self.auth_header, None)
        session_result = await self.session_manager.delete(token)
        return session_result

    async def me(self, token=None, user_id=None):
        if hasattr(self, "user_manager"):
            if self.user:
                me = self.user
            else:
                token = self.request.headers.get(self.auth_header, token)
                me = await self.user_manager.me(token, user_id)
                if me:
                    me = self.base_user_class(self.obj, me)
                    me["token"] = token
                self.user = me
        else:
            me = None
        return me

    async def update(self, push_token=None):
        token = self.request.headers.get(self.auth_header, None)
        ip = self.request.remote_ip
        # country = await location(ip)
        session = await self.session_manager.update(token, ip, location=None, push_token=push_token)
        return session



from datetime import datetime
from hashlib import sha1
from os import urandom

from peewee import DoesNotExist, IntegrityError

import cfg
from microservice import BasicObject
from microservice.exceptions import InternalError, ApiError
from microservice.functions import gravatar, send_mail
from microservice.managers.manager import DataManager
from microservice.models import User


class UserManager(DataManager):
    model = User

    async def create(self, user_data):
        """
        Берет юзера из базы, если такой id уже найден
        Используется для регистрации по емейлу/паролю
        """
        if not user_data.get("picture"):
            avatar = await gravatar(user_data["login"])
            user_data["picture"] = None
        user_data.update(
            password_hash=self.hash(user_data["password"]) if user_data["password"] else None,
            code=sha1(urandom(16)).hexdigest() if user_data["reg_method"] else None,
            registration_date=datetime.now(),
            status="unconfirmed",
        )
        user_obj, created = await self.obj.get_or_create(
            self.model,
            login=user_data["login"],
            defaults=user_data
        )
        if not created:
            raise InternalError("#auth #registration #field login already taken")
        user = user_obj.object()
        return user

    async def oauth(self, user_data):
        user_data.update(
            registration_date=datetime.now(),
            status="active",
        )
        login = user_data["login"]
        del user_data["login"]
        user_obj, created = await self.obj.get_or_create(
            self.model,
            login=login,
            defaults=user_data
        )
        user = user_obj.object()
        return user, created

    async def read(self, user_id=None, login=None, password=None, email=None, network=None):
        if user_id:
            user = await self.me(user_id=user_id)
            if not user:
                raise InternalError("#not_found user_id not found")
        elif login and password:
            user = await self.me(login=login)
            if not user:
                raise InternalError("#not_found login not found")
            if user["password_hash"] != self.hash(password):
                raise InternalError("#wrong_password Wrong password")
        elif login and network:
            user = await self.me(login=login)
            if not user:
                raise InternalError("#not_found login not found")
        elif email:
            user = await self.me(email=email)
            if not user:
                raise InternalError("#not_found email not found")
        else:
            raise InternalError("#missing #field login and password required")
        return user

    async def me(self, token=None, user_id=None, login=None, email=None):
        me_obj = await self.obj.execute(self.model.raw("""
            select u.*, row_to_json(s.*) "session"
            from "user" u
            LEFT JOIN "session" s ON u.id = s.user_id
            where (token = %s and s.expire > NOW()) OR u.id = %s OR u.login = %s OR u.email = %s
        """, token, user_id, login, email))
        me = self.extract_one(me_obj, extra_attrs=["session"])
        return me

    async def update(self, user_data):
        user_obj = await self.obj.get(self.model, id=user_data["id"])
        user = user_obj.object()
        if isinstance(user_data, BasicObject):
            user_data -= user  # prevent rewrite, just update

        if user_data.get("new_password"):
            user_data["password_hash"] = self.hash(user_data["new_password"])
        if user_data.get("email"):
            user_data["code"] = sha1(urandom(16)).hexdigest()
            user_data["status"] = "unconfirmed"
            user_data["login"] = user_data["email"]
            user_data["reg_method"] = "email"

        # изменение данных
        try:
            user.update(user_data)
            user_obj = self.model.from_dict(user, ignore_unknown=True)
            await self.obj.update(user_obj)
        except IntegrityError:
            if user_data.get("email"):
                raise InternalError("#email_not_available Error when updating user data")
            else:
                raise InternalError("#foreign_key_error Error when updating user data")

        return user.code

    async def generate_code(self, email):
        """
        Сгенерировать код для восстановления пароля
        :param email: емейл юзера
        """
        try:
            user_obj = await self.obj.get(self.model, email=email)
        except DoesNotExist:
            raise InternalError("#not_found email not found")
        else:
            user_obj.code = sha1(urandom(16)).hexdigest()
            await self.obj.update(user_obj)
        return user_obj.object()

    async def confirm(self, code, password=None):
        """
        Активировать аккаунт
        :param code: код активации
        :param password: пароль, если необходимо сменить
        """
        try:
            user_obj = await self.obj.get(self.model, code=code)
        except DoesNotExist:
            raise InternalError("#wrong_code Wrong code")
        else:
            user_obj.code = None
            if password:
                user_obj.password_hash = self.hash(password)
            user_obj.status = "confirmed"
            await self.obj.update(user_obj)

    async def check(self, login=None, email=None):
        """
        Проверить досутпен ли такой email
        """
        try:
            if login:
                await self.obj.get(self.model, login=login)
            elif email:
                await self.obj.get(self.model, email=email)
            else:
                raise InternalError("Something wrong")
        except DoesNotExist:
            return "not_found"
        else:
            return "found"

    def hash(self, password):
        """
        сгенерировать хеш
        :param password: пароль
        :return: хеш пароля (str)
        """
        if password:
            return sha1(str(password).encode("utf-8")).hexdigest()
        else:
            return None



from prometheus_client import Gauge, Counter


class Metrics:
    requests_per_node = Counter("requests_per_node", "Number of requests per second per node")
    requests_total = Counter("requests_total", "Number of requests per second * number of nodes")
    latency = Gauge("latency", "Total time of request processed")
    transactions = Gauge("transactions", "Number of transactions")
    errors_4xx = Counter("errors_4xx", "Number of http errors 4xx")



from microservice.handlers.container import Data
from microservice.handlers.handler import BasicHandler
from microservice.managers.manager import DataManager
from microservice.middleware.session import check
from microservice.middleware.objects import BasicObject, Collection
from microservice.server import Server



from setuptools import setup, find_packages

setup(
    name="microservice",
    version="0.3.24",
    description="Powerful REST API microservice on Tornado",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/pasaranax/microservice",
    author="Mikhail Bulygin",
    author_email="pasaranax@gmail.com",
    packages=find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)'
    ],
    install_requires=[
        "aiopg>=0.13.2,<1",
        "oauth2>=1.9.0.post1,<1.10",
        "peewee>=2.10.2,<3",
        "peewee-async>=0.5.12,<0.6",
        "prometheus-client>=0.3.0,<1",
        "pyTelegramBotAPI>=3.6.5,<3.7",
        "raven>=6.9.0,<6.10",
        "requests>=2.19,<3",
        "tornado>=5.1,<6",
        "ua-parser>=0.8.0,<1",
        "urllib3>=1.22,<2",
        "google-auth>=1.5.0,<1.6",
        "pytz>=2018.5",
        "aioredis>=1.1.0,<2",
        "asgiref>=2.3.2,<2.4"
    ]
)



import cfg
from peewee import Model, CharField, DoubleField, ForeignKeyField, PrimaryKeyField, DateTimeField, IntegerField, SQL
from peewee_asyncext import PooledPostgresqlExtDatabase
from playhouse.postgres_ext import JSONField
from playhouse.shortcuts import model_to_dict, dict_to_model

from microservice.middleware.objects import BasicObject

connection = PooledPostgresqlExtDatabase(
    cfg.db.database,
    host=cfg.db.host,
    port=cfg.db.port,
    user=cfg.db.user,
    password=cfg.db.password,
    register_hstore=False
)


class BasicModel(Model):
    id = PrimaryKeyField()

    class Meta:
        database = connection

    def __str__(self):
        if self.get("name"):
            name = self.name
        else:
            name = self.id

        return "<{} {}>".format(self.__class__.__name__, name)

    def dict(self, recurse=False, **kwargs):
        """
        Преобразовать peewee объект в словарь
        :return: dict
        """
        body = model_to_dict(self, recurse, **kwargs)
        return body

    @classmethod
    def from_dict(cls, d, ignore_unknown=True):
        """
        Собрать модель из словаря
        :param d: словарь с полями соответствующими модели
        :param ignore_unknown: игнорировать поля словаря, которых нет у модели
        :return: модель
        """
        model = dict_to_model(cls, d, ignore_unknown=ignore_unknown)
        return model

    def object(self, recurse=False, **kwargs):
        obj = BasicObject(model_to_dict(self, recurse, **kwargs))
        obj.model = self.__class__
        return obj


class Migrations(BasicModel):
    name = CharField()
    time = DateTimeField(default=SQL("NOW()"))


class User(BasicModel):
    """
    Пользователь
    """
    login = CharField(unique=True)
    email = CharField(null=True, unique=True)
    phone = CharField(null=True)
    password_hash = CharField(null=True)
    first_name = CharField(default="")
    last_name = CharField(null=True)
    role = CharField(null=True, default="user")
    picture = JSONField(null=True)
    status = CharField(default="unconfirmed")
    code = CharField(null=True)
    rating = DoubleField(default=0)
    reg_method = CharField(null=True)
    registration_date = DateTimeField(null=True)
    money = IntegerField(default=0)
    lang = CharField(default="en")
    timezone = CharField(default="UTC")
    utm_source = CharField(max_length=1000, null=True)
    utm_medium = CharField(max_length=1000, null=True)
    utm_campaign = CharField(max_length=1000, null=True)

    def __str__(self):
        return "{id}: {first_name} {last_name} ({pair})".format(**self.dict())


class Social(BasicModel):
    """
    Подключенные социальные сети
    """
    user = ForeignKeyField(User, on_delete="CASCADE")
    network = CharField(null=True)
    social_id = CharField(null=True, max_length=1000)
    access_token = CharField(null=True, max_length=1000)


class Session(BasicModel):
    """
    Сессия пользователя
    """
    user = ForeignKeyField(User, on_delete="CASCADE")
    token = CharField(null=True, index=True)
    rand = CharField(null=True)
    expire = DateTimeField(null=True)
    client = CharField(null=True)
    os = CharField(null=True)
    user_agent = CharField(null=True)
    login_method = CharField(null=True)
    push_token = CharField(null=True, max_length=2000)
    ip = CharField(null=True)
    location = JSONField(null=True)
    last_login = DateTimeField(null=True)



import logging

from microservice.middleware.objects import BasicObject


class BaseUser(BasicObject):
    """
    Dummy user, this object is created every time after successful @check(anonymous=False)
    """
    user_manager_class = None

    def __init__(self, obj, user_dict=None):
        super(BaseUser, self).__init__(user_dict)
        self.obj = obj
        self.user_dict = user_dict
        if obj:
            if not self.user_manager_class:
                from microservice.managers.user import UserManager
                self.user_manager_class = UserManager
            else:
                self.user_manager = self.user_manager_class(obj)

    async def save(self):
        await self.user_manager.update(self.data)

    async def reincarnate(self, user_id):
        result = await self.user_manager.me(user_id=user_id)
        self.__init__(self.obj, result)

    async def check_self(self, handler):
        return True

    def capture_message(self, message, extra=None, level=logging.WARNING):
        if hasattr(self, "sentry_client"):
            self.sentry_client.user_context(dict(self))
            self.sentry_client.captureMessage(message, extra=extra, level=level)

    def capture_exception(self, extra=None):
        if hasattr(self, "sentry_client"):
            self.sentry_client.user_context(dict(self))
            self.sentry_client.captureException(extra=extra)



try:
    from handlers.metrics import MetricsHandler
except ImportError:
    from microservice.handlers.metrics import MetricsHandler
from microservice.handlers.handler import BasicHandler


class Router:
    def __init__(self, handlers):
        """
        формат handlers: (endpoint str, {version int: handler handler})
        """
        self.handlers = handlers
        self.min_version = 1  # минимальная поддерживаемая версия
        self.latest = 99  # latest actual api version
        self._routes = []

    def gen_url(self, v, endpoint, handler):
        min_v = sorted(handler.keys())[0]
        if v in handler:
            choosen_handler = handler[v]
            choosen_handler.version = v
        else:
            if v > min_v:
                choosen_handler = self.gen_url(v-1, endpoint, handler)[1]
            else:
                return None

        return '/v{}/{}'.format(v, endpoint), choosen_handler

    @property
    def routes(self):
        for endpoint, handler in self.handlers:
            for v in range(self.min_version, self.latest+1):
                route = self.gen_url(v, endpoint, handler)
                if route:
                    self._routes.append(route)

        #  Служебные эндпоинты
        self._routes.extend([
            ("/metrics", MetricsHandler),
            (".*", BasicHandler),
        ])
        return self._routes



import binascii
import gzip
import json
import logging
import re
import traceback
from hashlib import sha1
from base64 import b64decode
from concurrent.futures import ThreadPoolExecutor
from json import JSONDecodeError
from time import perf_counter as pc
from urllib.parse import unquote, parse_qs

from raven.contrib.tornado import SentryMixin
from telebot.apihelper import ApiException
from tornado.escape import json_decode
from tornado.netutil import is_valid_ip
from tornado.web import RequestHandler

import cfg
from microservice.exceptions import ApiError
from microservice.functions import TelegramReporter
from microservice.metrics import Metrics
from microservice.middleware.session import Session
from microservice.middleware.cache import RedisCache


class Answer:
    """
    Ответ, отправляемый клиенту
    """

    def __init__(self):
        """
        Создает ответ на основе словаря data
        """
        self.t = pc() * 1000
        self.answer = {}
        self.raw = None

    def compose(self, message=None, data=None, error=None, meta=None, raw=None):
        """
        Необходимо вызвать для составления ответа
        """
        assert message is not None or error is not None or raw is not None
        self.raw = raw
        self.answer = {"message": message, "error": error, "data": data, "meta": meta}

    def dict(self):
        """
        Возвращает словарь с данными
        """
        self.answer["time"] = self.mark_time()
        return self.answer

    def gzip(self):
        """
        gzipped json
        """
        self.answer["time"] = self.mark_time()
        return gzip.compress(json.dumps(self.answer).encode("utf-8"), compresslevel=cfg.app.gzip_output)

    def raw_data(self):
        self.answer["time"] = self.mark_time()
        return self.raw

    def mark_time(self):
        """
        Замерить время, поставить временную отметку между текущим временем и предыдущей отметкой
        """
        Metrics.latency.set(pc() * 1000 - self.t)
        return round(pc() * 1000 - self.t, 3)


class SentryMixinExt(SentryMixin):
    def get_sentry_user_info(self):
        """
        Data for sentry.interfaces.User

        Default implementation only sends `is_authenticated` by checking if
        `tornado.web.RequestHandler.get_current_user` tests postitively for on
        Truth calue testing
        """
        try:
            user = self.get_current_user()
        except Exception:
            return {}
        return {
            'user': user
        }


class BasicHandler(SentryMixinExt, RequestHandler):
    executor = ThreadPoolExecutor(max_workers=cfg.app.max_workers_on_executor)
    session_class = None
    cache_method = None  # "user", "all", None
    cache_lifetime = 10
    atomize = True  # every transaction creates atomic transaction in db

    def get_session_class(self):
        return self.session_class or Session

    def initialize(self):
        if self.application.redis_connection:
            self.cache = RedisCache(self.application.redis_connection)
        else:
            self.cache = None
            self.cache_method = None
        self.cached = False
        self.answer = Answer()
        forwarded_ip = self.request.headers.get("X-Forwarded-For", self.request.remote_ip)
        if is_valid_ip(forwarded_ip):
            self.request.remote_ip = forwarded_ip
        self.loop = self.application.loop
        self.json_body = None
        self.session = self.get_session_class()(self.request, self.body, self.args, self.application.objects)
        self.queue = []
        self.version = self.version if hasattr(self, "version") else 0
        self.finished = False

    def init(self):
        """Redefine this method instead of self.initialize()"""
        pass

    async def prepare(self, *kwargs):
        """
        Выполняется перед  обработкой запроса
        """
        self.add_header('Access-Control-Allow-Origin', '*')
        self.add_header('Access-Control-Allow-Headers', 'Content-Type, %s, X-Client-Type, Origin, Cache-Control, '
                                                        'X-Requested-With, Accept' % self.get_session_class().auth_header)
        self.add_header('Access-Control-Allow-Methods', 'GET, POST, PATCH, PUT, DELETE, OPTIONS, HEAD')
        self.add_header('Access-Control-Allow-Credentials', 'true')
        self.set_header('Content-Type', 'application/json; charset=UTF-8')
        try:
            self.request.headers["User-Agent"] = b64decode(self.request.headers.get("User-Agent", "")).decode("utf-8")
        except (binascii.Error, UnicodeDecodeError):
            pass
        self.client_platform = self.request.headers.get("X-Platform", "")
        self.client_version = self.request.headers.get("X-Version", "")
        try:
            self.api_version = int(self.request.uri.split("/")[1][1:])
        except ValueError:
            self.api_version = 0
        self.endpoint = re.sub("\d", "", "/".join(self.request.uri.split("?")[0].split("/")[2:])).rstrip("/")

        # Caching: if cache found don't call method, just return cached answer
        if self.cache and self.request.headers.get("Cache-Control") != "no-cache" and self.cache_method:
            hash_exists = await self.cache.check_request(self.request_hash(for_user=self.cache_method == "user"))
            if hash_exists:
                restored_answer = await self.cache.restore_answer(self.request_hash(for_user=self.cache_method == "user"))
                if restored_answer == "":
                    self.compose(error="#duplicate_request Request in progress", send=True)
                else:
                    self.answer.answer = restored_answer
                    self.cached = True
                    self.send_result()
            else:
                await self.cache.store_request(self.request_hash(for_user=self.cache_method == "user"), self.cache_lifetime, "")
        self.init()

    async def get(self, *args, **kwargs):
        self.compose(error="#method_not_allowed", status=405, send=True)

    async def post(self, *args, **kwargs):
        self.compose(error="#method_not_allowed", status=405, send=True)

    async def delete(self, *args, **kwargs):
        self.compose(error="#method_not_allowed", status=405, send=True)

    async def patch(self, *args, **kwargs):
        self.compose(error="#method_not_allowed", status=405, send=True)

    async def put(self, *args, **kwargs):
        self.compose(error="#method_not_allowed", status=405, send=True)

    def head(self, *args, **kwargs):
        self.compose(error="#method_not_allowed", status=405, send=True)

    def options(self, *args, **kwargs):
        self.set_status(204)
        self.finish()

    def compose(self, message=None, result=None, error=None, status=None, send=False, raw=None):
        """
        Сформировать ответ
        :param message: сообщение, которое выводится в случае отсутствия ошибки
        :param result: должен содержать параметры data или error
        :param error: текст ошибки
        :param status: кастомный статус код
        :param send: сразу отправить результат клиенту
        """
        if error or (result and result.error):
            error = str(error) if error else None or result.error
            code, *error_message = error.split(maxsplit=1)
            error_message = "".join(error_message)
            if not code or not code.startswith("#"):
                code = "#error"
                error_message = error
            logging.warning("Error: '{}'. Body: {}".format(error, self.body()))
            # Статус из хештегов
            if not status:
                if code in ("#auth", "#oauth", "#access_denied", "#wrong_password"):
                    status = 401
                elif code == "#not_found":
                    status = 404
                elif code in ("#wrong_format", "#wrong_type", "#api_error"):
                    status = 417
                elif code == "#method_not_allowed":
                    status = 405
            self.set_status(status or 400)
            self.answer.compose(message=None, error={"code": code, "message": error_message})
            Metrics.errors_4xx.inc(1)
            self.captureMessage(error, level=logging.WARNING, extra={
                "cached": self.cached,
                "log": self.make_log()
            })
        else:
            self.set_status(status or 200)
            if result:
                result.meta.update(result.kwargs)
                self.answer.compose(message, data=result.data, meta=result.meta)
            elif raw:
                self.answer.compose(raw=raw)
            else:
                self.answer.compose(message)
        if send:
            self.send_result()

    def send_result(self):
        if not self.finished:
            if self.answer.raw:
                self.finish(self.answer.raw_data())
            elif cfg.app.gzip_output > 0:
                self.set_header('Content-Encoding', 'gzip')
                self.finish(self.answer.gzip())
            else:
                self.finish(self.answer.dict())
            self.finished = True

    def make_log(self):
        log_record = "{}: ({}) {} {} [{}] {} ms".format(
            self.get_status(), self.request.remote_ip, self.request.method, unquote(self.request.uri),
            self.request.headers.get("User-Agent", ""), self.answer.mark_time()
        )
        if self.session.user is not None:
            log_record = "{} (token: {}, {}: {} {})".format(
                log_record, self.request.headers.get("X-Token"),
                self.session.user["id"], self.session.user["first_name"] or self.session.user["email"], self.session.user["last_name"] or "\b",
            )
        elif self.request.headers.get("X-Token"):
            log_record = "{} (token: {} {})".format(
                log_record, self.request.headers.get("X-Token"), "From cache" if self.cached else "\b"
            )
        return log_record

    def on_finish(self):
        """
        Выполняется после отправки ответа
        """
        log_record = self.make_log()
        if str(self.get_status()).startswith("2"):
            logging.info(log_record)
        elif str(self.get_status()).startswith("4"):
            logging.warning(log_record)
        elif str(self.get_status()).startswith("5"):
            logging.error(log_record)
        else:
            logging.info(log_record)
        Metrics.requests_per_node.inc(1)
        Metrics.requests_total.inc(cfg.app.number_of_nodes)

    def get_current_user(self):
        return dict(self.session.user)

    def valid(self, name, value, default=None, coerce=None, check=None, error=""):
        """
        Проверить значение на валидность
        :param error: текст в случае ошибки
        :param value: значение
        :param default: если значение None, то взять default
        :param coerce: привести к типу или применить функцию
        :param check: проверить функцией (должна вернуть True)
        :return: возвращает значение
        """
        if value is None:
            value = default
        if coerce and value is not None:
            try:
                value = coerce(value)
            except ValueError:
                raise ApiError("#wrong_type {}, expected {}. {}".format(name, coerce, error))
        if check and value is not None and not check(value):
            raise ApiError("#wrong_format {}. {}".format(name, error))
        return value

    def body(self, name=None, default=None, required=False, coerce=None, check=None, error=""):
        """
        Получить параметр из тела запроса с проверками
        :param name: имя параметра
        :param default: значение по умолчанию, если не нет в теле
        :param required: является ли обязательным
        :param coerce: привести к типу (можно применить функцию)
        :param check: функция проверки, должна вернуть True
        :return: возвращает значение параметра
        """
        if self.request.body:
            if self.json_body is None:
                try:
                    self.json_body = json_decode(self.request.body)
                    if not isinstance(self.json_body, (dict, list)):
                        raise JSONDecodeError("Body must be dict or list", self.json_body, 0)
                except (JSONDecodeError, UnicodeDecodeError) as e:
                    self.json_body = dict()
                    for key, value in parse_qs(self.request.body.decode("utf-8")).items():
                        self.json_body[key] = value[0] if len(value) > 0 else value
                    # raise ApiError("#json #parse {}".format(e))
        else:
            self.json_body = dict()

        if name:
            value = self.json_body.get(name, default)
            if required and value in (None, ""):
                # self.compose(error="#missing #field {}".format(name))
                raise ApiError("#missing #field '{}' {}".format(name, error))
            value = self.valid(name, value, default, coerce, check, error)
            return value
        else:
            return self.json_body

    def args(self, name, default=None, required=False, coerce=None, check=None, error=""):
        value = self.get_query_argument(name, default)
        if required and value in (None, ""):
            # self.compose(error="#missing #argument {}".format(name))
            raise ApiError("#missing #argument '{}' {}".format(name, error))
        value = self.valid(name, value, default, coerce, check, error)
        return value

    def write_error(self, status_code, **kwargs):
        data = dict()
        self.set_header('Content-Type', 'application/json')
        if "exc_info" in kwargs:
            data["traceback"] = list(traceback.format_exception(*kwargs["exc_info"]))
            info = {
                "log": self.make_log(),
                "token": self.request.headers.get(self.get_session_class().auth_header, "none"),
                "body": json.dumps(self.body(), sort_keys=True, indent=4, separators=(',', ': '))
            }
            data["error"] = data["traceback"][-1]
            traceback.print_exc()
            if cfg.app.send_telegram_errors:
                try:
                    TelegramReporter.telegram_error(data["traceback"], info)
                except ApiException as e:
                    try:
                        TelegramReporter.telegram_error(data["traceback"][-5:], info)
                    except ApiException as e:
                        info["body"] = "Too big body (truncated)"
                        logging.warning("Can't send telegram report: {}".format(e))
                        TelegramReporter.telegram_error(data["traceback"][-5:], info)
        if cfg.app.debug:
            self.finish(data)
        else:
            self.finish()

    def filter(self, d, allowed):
        """
        Отфильтровать словарь на допустимые значения
        :param d: словарь входящих значений
        :param allowed: список допустимых полей
        :return: dict
        """
        return {k: v for k, v in d.items() if k in allowed}

    def drop_nones(self, d, required=None):
        return {k: v for k, v in d.items() if v is not None or (required and k in required)}

    def request_hash(self, for_user=False):
        s = "{}.{}.{}.{}.{}".format(
            self.request.method,
            self.request.uri,
            self.request.headers.get("X-Token", "") if for_user else "",
            self.request.body,
            cfg.app.secret_key
        )
        hash = sha1(s.encode("utf-8"))
        return hash.hexdigest()



import itertools
import logging
import os
import platform
import socket
from email.mime.text import MIMEText
from hashlib import md5
from smtplib import SMTP_SSL
from time import perf_counter as pc

import requests
from asgiref.sync import sync_to_async
from telebot import TeleBot

import cfg


class TelegramReporter:
    last_error = None
    bot = TeleBot(cfg.app.telegram_bot_token)
    host = socket.gethostbyname_ex(socket.gethostname())
    addr = os.environ.get('REMOTE_ADDR', "")

    @staticmethod
    def telegram_error(traceback, info=None):
        """
        Отправляет сообщение в телеграм с ошибкой на сервере
        :param traceback: list of error lines
        """
        config = cfg.app.telegram_reporter.get(platform.node())
        if not config:
            logging.warning("Telegram reporter: Unknown host, can't send message")
        else:
            text = (
                "<b>Error on {name}, node: {host}, {addr}</b>\n"
                "{log}\n"
                "<pre>{body}</pre>\n"
                "<b>Token:</b> {token}\n"
                "<pre>{traceback}</pre>"
            ).format(
                name=config["name"],
                host=TelegramReporter.host,
                addr=TelegramReporter.addr,
                log=info["log"],
                body=info["body"],
                token=info["token"],
                traceback="".join(traceback).replace("<", "").replace(">", ""),
            )
            logging.info("Sending error from node {} to chat {}".format(config["name"], config["chat_id"]))
            if text != TelegramReporter.last_error:
                TelegramReporter.last_error = text
                TelegramReporter.bot.send_message(config["chat_id"], text, parse_mode="html")

    @staticmethod
    def send_message(chat_id, message, only_prod=False, ignore_tests=True, footer=True):
        if ignore_tests and TelegramReporter.host[0].startswith("runner"):
            return
        if only_prod and "Production" not in TelegramReporter.host[0]:
            return
        if footer:
            message = (
                "{}\n"
                "<b>Host:</b> {}, {} "
                "<b>DB:</b> {} "
            ).format(message, TelegramReporter.host, TelegramReporter.addr, cfg.db.host)
        TelegramReporter.bot.send_message(chat_id, message, disable_web_page_preview=True, parse_mode="html")

    @staticmethod
    @sync_to_async
    def send_async(chat_id, message, only_prod=False, ignore_tests=True, footer=False):
        TelegramReporter.send_message(chat_id, message, only_prod=only_prod, ignore_tests=ignore_tests, footer=footer)

    @staticmethod
    def send_picture(chat_id, link):
        TelegramReporter.bot.send_photo(chat_id, link)


def send_mail(to, text, subject, type="plain"):
    smtp = SMTP_SSL(cfg.app.smtp_host, port=465)
    smtp.login(cfg.app.smtp_login, cfg.app.smtp_password)

    msg = MIMEText(text or "", _subtype=type)
    msg["Subject"] = subject
    msg["From"] = cfg.app.smtp_from
    msg["To"] = to

    smtp.sendmail(cfg.app.support_email, to, msg.as_string())
    smtp.quit()


def chunks(iterable, n):
    """
    Разделить итератор на куски по n элементов
    :param iterable: итератор
    :param n: количество элементов в куске
    """
    it = iterable
    while True:
        chunk = tuple(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk


# def extract_results(results, *args, **kwargs):
#     """
#     Извлечь список из результатов пиви или эластика
#     :param extra_attrs: аттрибуты из результата, которые надо включить в словарь (peewee)
#     :param results: ответ эластика или пиви
#     :return: list
#     """
#     if len(results) > 0:
#         lst = [x.dict(*args, **kwargs) for x in results]
#     else:
#         lst = []
#     return lst
#
#
# def extract_one(results, *args, **kwargs):
#     for x in results:
#         return x.dict(*args, **kwargs)
#     return None


async def location(ip):
    country = requests.request("GET", "http://api.ipstack.com/{}?access_key={}".format(ip, cfg.app.ipstack_api_key)).json()
    return country


async def gravatar(email):
    digest = md5(email.lower().encode("utf-8")).hexdigest()
    return "https://www.gravatar.com/avatar/{}?d=retro&s={}".format(digest, 500)


class DummyAtomic:
    async def __aenter__(self):
        return self

    async def __aexit__(self, *kwargs):
        pass


def check_atomic(obj, atomize=True):
    if obj and atomize is True:
        return obj.atomic()
    else:
        return DummyAtomic()


class Timer:
    def __init__(self):
        self.start_time = pc()
        self.mark = pc()
        self.counter = 0

    def step(self, step="Step"):
        from_prev = pc() - self.mark
        from_start = pc() - self.start_time
        logging.debug("Step {} {}: {:.3f} from prev step, {:.3f} from start".format(self.counter, step, from_prev*1000, from_start*1000))
        self.mark = pc()
        self.counter += 1
